# Copyright (c) 2019, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import copy
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle

import rospkg

from ._version import __version__
from .cache_tools import compute_filename_hash
from .cache_tools import write_cache_file
from .cache_tools import PICKLE_CACHE_EXT

"""
Rosdep needs to store data that isn't used to resolve rosdep keys, but needs to be cached during
`rosdep update`.
"""

META_CACHE_DIR = 'meta.cache'


def get_meta_cache_dir():
    """Return storage location for cached meta data."""
    ros_home = rospkg.get_ros_home()
    return os.path.join(ros_home, 'rosdep', META_CACHE_DIR)


class CacheWrapper(object):
    """Make it possible to introspect cache in case some future bug needs to be worked around."""

    def __init__(self, category, data):
        # The version of rosdep that wrote the category
        self.rosdep_version = __version__
        # The un-hashed name of the category
        self.category_name = category
        # The stuff being stored
        self.data = data

    @property
    def data(self):
        # If cached data type is mutable, don't allow modifications to what's been loaded
        return copy.deepcopy(self.__data)

    @data.setter
    def data(self, value):
        self.__data = copy.deepcopy(value)


class MetaDatabase:
    """
    Store and retrieve metadata from rosdep cache.

    This data is fetched during `rosdep update`, but is not a source for resolving rosdep keys.
    """

    def __init__(self, cache_dir=None):
        if cache_dir is None:
            cache_dir = get_meta_cache_dir()

        self._cache_dir = cache_dir
        self._loaded = {}

    def set(self, category, metadata):  # noqa: A003
        """Add or overwrite metadata in the cache."""
        wrapper = CacheWrapper(category, metadata)
        # print(category, metadata)
        write_cache_file(self._cache_dir, category, wrapper)
        self._loaded[category] = wrapper

    def get(self, category, default=None):
        """Return metadata in the cache, or None if there is no cache entry."""
        if category not in self._loaded:
            self._load_from_cache(category, self._cache_dir)

        if category in self._loaded:
            return self._loaded[category].data

        return default

    def _load_from_cache(self, category, cache_dir):
        filename = compute_filename_hash(category) + PICKLE_CACHE_EXT
        try:
            with open(os.path.join(self._cache_dir, filename), 'rb') as cache_file:
                self._loaded[category] = pickle.loads(cache_file.read())
        except FileNotFoundError:
            pass



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Ken Conley/kwc@willowgarage.com

import os
import traceback
from unittest.mock import Mock, patch, call


def get_test_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'debian'))


def test_dpkg_detect():
    from rosdep2.platforms.debian import dpkg_detect

    with patch('rosdep2.platforms.debian.read_stdout') as mock_read_stdout:
        mock_read_stdout.side_effect = [('', ''), '']
        val = dpkg_detect([])
        assert val == [], val
        assert mock_read_stdout.call_count == 2
        assert mock_read_stdout.call_args_list[0] == call(['dpkg-query', '-W', "-f='${Package} ${Status}\n'"], True)
        assert mock_read_stdout.call_args_list[1] == call(['apt-cache', 'showpkg'])

    with patch('rosdep2.platforms.debian.read_stdout') as mock_read_stdout:
        mock_read_stdout.side_effect = [('', ''), '']
        val = dpkg_detect(['tinyxml-dev'])
        assert val == [], val
        assert mock_read_stdout.call_count == 2
        assert mock_read_stdout.call_args_list[0] == call(['dpkg-query', '-W', "-f='${Package} ${Status}\n'",
                                                           'tinyxml-dev'], True)
        assert mock_read_stdout.call_args_list[1] == call(['apt-cache', 'showpkg', 'tinyxml-dev'])

    with open(os.path.join(get_test_dir(), 'dpkg-python-apt'), 'r') as f:
        dpkg_python_apt_test_content = f.read()

    with patch('rosdep2.platforms.debian.read_stdout') as mock_read_stdout:
        mock_read_stdout.side_effect = [(dpkg_python_apt_test_content, ''), '']
        val = dpkg_detect(['apt', 'tinyxml-dev', 'python'])
        assert val == ['apt', 'python'], val
        assert mock_read_stdout.call_count == 2
        print(mock_read_stdout.call_args_list[0])
        assert mock_read_stdout.call_args_list[1] == call(['apt-cache', 'showpkg', 'tinyxml-dev'])

    # test version lock code (should be filtered out w/o validation)
    with patch('rosdep2.platforms.debian.read_stdout') as mock_read_stdout:
        mock_read_stdout.side_effect = [(dpkg_python_apt_test_content, ''), '']
        val = dpkg_detect(['apt=1.8', 'tinyxml-dev', 'python=2.7'])
        assert val == ['apt=1.8', 'python=2.7'], val
        assert mock_read_stdout.call_count == 2
        assert mock_read_stdout.call_args_list[1] == call(['apt-cache', 'showpkg', 'tinyxml-dev'])


def test_read_apt_cache_showpkg():
    from rosdep2.platforms.debian import _read_apt_cache_showpkg

    m = Mock()
    with open(os.path.join(get_test_dir(), 'showpkg-curl-wget-libcurl-dev'), 'r') as f:
        m.return_value = f.read()
    pkgs = ['curl', 'wget', '_not_existing', 'libcurl-dev', 'ros-kinetic-rc-genicam-api']
    results = list(_read_apt_cache_showpkg(pkgs, exec_fn=m))
    assert len(results) == len(pkgs), results

    package, virtual, providers = results[0]
    assert package == 'curl', package
    assert not virtual
    assert providers is None, providers

    package, virtual, providers = results[1]
    assert package == 'wget', package
    assert not virtual
    assert providers is None, providers

    package, virtual, providers = results[2]
    assert package == '_not_existing', package
    assert not virtual
    assert providers is None, providers

    package, virtual, providers = results[3]
    assert package == 'libcurl-dev', package
    assert virtual, providers

    package, virtual, providers = results[4]
    assert package == 'ros-kinetic-rc-genicam-api', package
    assert not virtual
    assert providers is None, providers


def test_AptInstaller():
    from rosdep2.platforms.debian import AptInstaller

    @patch('rosdep2.platforms.debian.read_stdout')
    @patch.object(AptInstaller, 'get_packages_to_install')
    def test(expected_prefix, mock_get_packages_to_install, mock_read_stdout):
        installer = AptInstaller()
        mock_get_packages_to_install.return_value = []
        mock_read_stdout.return_value = ''

        mock_get_packages_to_install.return_value = ['a', 'b']
        expected = [expected_prefix + ['apt-get', 'install', '-y', '--no-install-recommends', 'whatever']]
        val = installer.get_install_command(['whatever'], interactive=False)
        assert val == expected, val
        expected = [expected_prefix + ['apt-get', 'install', 'whatever']]
        val = installer.get_install_command(['whatever'], interactive=True)
        assert val == expected, val

        # oneshot
        expected = [expected_prefix + ['apt-get', 'install', '-y', '--no-install-recommends', 'whatever']]
        val = installer.get_install_command(['whatever'], interactive=False, oneshot=["apt"])
        print('VAL', val)
        assert val == expected, val
        expected = [expected_prefix + ['apt-get', 'install', 'whatever']]
        val = installer.get_install_command(['whatever'], interactive=True, oneshot=["apt"])
        assert val == expected, val
    try:
        if hasattr(os, 'geteuid'):
            with patch('rosdep2.installers.os.geteuid', return_value=1):
                test(['sudo', '-H'])
            with patch('rosdep2.installers.os.geteuid', return_value=0):
                test([])
        else:
            test([])
    except AssertionError:
        traceback.print_exc()
        raise



# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import hashlib
import os
import tempfile

from .core import CachePermissionError

try:
    import cPickle as pickle
except ImportError:
    import pickle

PICKLE_CACHE_EXT = '.pickle'


def compute_filename_hash(key_filenames):
    sha_hash = hashlib.sha1()
    if isinstance(key_filenames, list):
        for key in key_filenames:
            sha_hash.update(key.encode())
    else:
        sha_hash.update(key_filenames.encode())
    return sha_hash.hexdigest()


def write_cache_file(source_cache_d, key_filenames, rosdep_data):
    """
    :param source_cache_d: directory to write cache file to
    :param key_filenames: filename (or list of filenames) to be used in hashing
    :param rosdep_data: dictionary of data to serialize as YAML
    :returns: name of file where cache is stored
    :raises: :exc:`OSError` if cannot write to cache file/directory
    :raises: :exc:`IOError` if cannot write to cache file/directory
    """
    if not os.path.exists(source_cache_d):
        os.makedirs(source_cache_d)
    key_hash = compute_filename_hash(key_filenames)
    filepath = os.path.join(source_cache_d, key_hash)
    try:
        write_atomic(filepath + PICKLE_CACHE_EXT, pickle.dumps(rosdep_data, 2), True)
    except OSError as e:
        raise CachePermissionError('Failed to write cache file: ' + str(e))
    try:
        os.unlink(filepath)
    except OSError:
        pass
    return filepath


def write_atomic(filepath, data, binary=False):
    # write data to new file
    fd, filepath_tmp = tempfile.mkstemp(prefix=os.path.basename(filepath) + '.tmp.', dir=os.path.dirname(filepath))

    if (binary):
        fmode = 'wb'
    else:
        fmode = 'w'

    with os.fdopen(fd, fmode) as f:
        f.write(data)
        f.close()

    try:
        # switch file atomically (if supported)
        os.rename(filepath_tmp, filepath)
    except OSError:
        # fall back to non-atomic operation
        try:
            os.unlink(filepath)
        except OSError:
            pass
        try:
            os.rename(filepath_tmp, filepath)
        except OSError:
            os.unlink(filepath_tmp)



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Ken Conley/kwc@willowgarage.com

import os
import pytest
import yaml

from .conftest import path_to_url


rep122_install_command = """#!/bin/bash
set -o errexit
mkdir -p build
cd build
cmake ..
make
echo "About to run checkinstall make install"
sudo checkinstall -y --nodoc --pkgname=yaml-cpp-sourcedep make install
"""

rep122_check_presence_command = """#!/bin/bash
dpkg-query -W -f='${Package} ${Status}\\n' yaml-cpp-sourcedep | awk '{\\
if ($4 =="installed")
  exit 0
else
  print "yaml-cpp-sourcedep not installed"
  exit 1}'
"""

REP112_MD5SUM = '57cb9faf930e9c4f0822be8b27798248'


def get_test_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'source'))


def _subtest_rep112_rdmanifest(resolved):
    test_dir = get_test_dir()
    path = os.path.join(test_dir, 'rep112-example.rdmanifest')
    manifest = yaml.safe_load(open(path))

    assert resolved.manifest == manifest
    assert resolved.manifest_url == path
    assert resolved.install_command == rep122_install_command
    assert resolved.check_presence_command == rep122_check_presence_command

    assert len(resolved.check_presence_command) == len(rep122_check_presence_command), '%s %s' % (len(resolved.check_presence_command), len(rep122_check_presence_command))

    assert resolved.exec_path == 'yaml-cpp-0.2.5'
    assert resolved.tarball == 'http://download.ros.org/downloads/yaml-cpp-0.2.5.tar.gz'
    assert resolved.alternate_tarball is None
    assert resolved.tarball_md5sum == 'b17dc36055cd2259c88b2602601415d9'


def test_SourceInstall():
    from rosdep2.platforms.source import InvalidRdmanifest, SourceInstall

    # tripwire
    SourceInstall()

    # test unpacking of dict
    manifest = {
        'md5sum': 'fake-md5',
        'exec-path': '/path',
        'install-script': 'echo hello',
        'check-presence-script': 'hello there',
        'uri': 'http://ros.org/',
        'alternate-uri': 'http://turtlebot.com/',
        'depends': ['foo', 'bar'],
    }
    resolved = SourceInstall.from_manifest(manifest, 'fake-url')
    assert resolved.manifest == manifest
    assert resolved.manifest_url == 'fake-url'
    assert resolved.install_command == 'echo hello'
    assert resolved.check_presence_command == 'hello there'
    assert resolved.exec_path == '/path'
    assert resolved.tarball == 'http://ros.org/'
    assert resolved.alternate_tarball == 'http://turtlebot.com/'
    assert resolved.tarball_md5sum == 'fake-md5'

    test_dir = get_test_dir()
    path = os.path.join(test_dir, 'rep112-example.rdmanifest')
    manifest = yaml.safe_load(open(path))
    resolved = SourceInstall.from_manifest(manifest, path)
    _subtest_rep112_rdmanifest(resolved)

    # TODO: test depends

    # test with bad dicts
    manifest = {
        'md5sum': 'fake-md5',
        'exec-path': '/path',
        'install-script': 'echo hello',
        'check-presence-script': 'hello there',
        'alternate-uri': 'http://turtlebot.com/',
        'depends': ['foo', 'bar'],
    }
    # uri is required
    try:
        SourceInstall.from_manifest(manifest, 'foo')
        assert False, 'should have raised'
    except InvalidRdmanifest as e:
        pass

    # test defaults
    manifest = dict(uri='http://ros.org/')
    resolved = SourceInstall.from_manifest(manifest, 'foo')
    assert resolved.exec_path == '.'
    assert resolved.install_command == ''
    assert resolved.check_presence_command == ''
    assert resolved.alternate_tarball is None
    assert resolved.tarball_md5sum is None


def test_is_installed():
    from rosdep2.platforms.source import SourceInstaller, SourceInstall
    resolved = SourceInstall()
    resolved.check_presence_command = """#!/bin/bash
exit 0
"""
    installer = SourceInstaller()
    assert installer.is_installed(resolved)


def test_source_detect():
    from rosdep2.platforms.source import source_detect, SourceInstall
    resolved = SourceInstall()
    resolved.check_presence_command = """#!/bin/bash
exit 0
"""
    assert [] == source_detect([])
    assert [resolved] == source_detect([resolved])

    def yes(*args, **kwds):
        return 0

    def no(*args, **kwds):
        return 1

    resolved = [SourceInstall(), SourceInstall(), SourceInstall(), SourceInstall()]
    for r in resolved:
        r.check_presence_command = ''

    retval = source_detect(resolved, exec_fn=yes)
    assert resolved == retval, retval
    assert [] == source_detect(resolved, exec_fn=no)


def test_SourceInstaller_get_install_command():
    from rosdep2.platforms.source import SourceInstaller, SourceInstall
    installer = SourceInstaller()

    resolved = SourceInstall()
    resolved.manifest_url = 'http://fake/foo'
    resolved.check_presence_command = """#!/bin/bash
exit 1
"""
    commands = installer.get_install_command([resolved])
    assert len(commands) == 1
    assert commands[0] == ['rosdep-source', 'install', 'http://fake/foo']

    resolved = SourceInstall()
    resolved.manifest_url = 'http://fake/foo'
    resolved.check_presence_command = """#!/bin/bash
exit 0
"""
    commands = installer.get_install_command([resolved])
    assert not commands


def test_SourceInstaller_resolve():
    from rosdep2.platforms.source import SourceInstaller, InvalidData
    test_dir = get_test_dir()

    url = path_to_url(os.path.join(test_dir, 'rep112-example.rdmanifest'))
    md5sum_good = REP112_MD5SUM
    md5sum_bad = 'fake'

    installer = SourceInstaller()
    try:
        installer.resolve({})
        assert False, 'should have raised'
    except InvalidData:
        pass
    try:
        installer.resolve(dict(uri=url, md5sum=md5sum_bad))
        assert False, 'should have raised'
    except InvalidData:
        pass
    resolved = installer.resolve(dict(uri=url, md5sum=md5sum_good))

    assert type(resolved) is list
    assert len(resolved) == 1
    # test for reinstall (to check the depends in rdmanifest)
    dependencies = installer.get_depends(dict(uri=url, md5sum=md5sum_good))
    assert dependencies == ['checkinstall'], 'Dependencies should resolve to checkinstall listed in the rdmanifest.'
    resolved = resolved[0]

    assert resolved.install_command == rep122_install_command
    assert resolved.check_presence_command == rep122_check_presence_command

    # test again to activate caching
    resolved = installer.resolve(dict(uri=url, md5sum=md5sum_good))
    assert type(resolved) is list, 'Cache should also return a list'
    assert len(resolved) == 1
    resolved = resolved[0]
    assert resolved.install_command == rep122_install_command
    assert resolved.check_presence_command == rep122_check_presence_command


def test_load_rdmanifest():
    from rosdep2.platforms.source import load_rdmanifest, InvalidRdmanifest
    # load_rdmanifest is just a YAML unmarshaller with an exception change
    assert 'str' == load_rdmanifest('str')
    assert {'a': 'b'} == load_rdmanifest('{a: b}')

    try:
        load_rdmanifest(';lkajsdf;klj ;l: a;kljdf;: asdf\n ;asdfl;kj')
        assert False, 'should have raised'
    except InvalidRdmanifest as e:
        pass


def test_get_file_hash():
    from rosdep2.platforms.source import get_file_hash
    path = os.path.join(get_test_dir(), 'rep112-example.rdmanifest')
    assert REP112_MD5SUM == get_file_hash(path)


def test_fetch_file():
    test_dir = get_test_dir()
    with open(os.path.join(test_dir, 'rep112-example.rdmanifest')) as f:
        expected = f.read()

    from rosdep2.platforms.source import fetch_file
    url = path_to_url(os.path.join(test_dir, 'rep112-example.rdmanifest'))
    contents, error = fetch_file(url, REP112_MD5SUM)
    assert not error
    assert contents == expected

    contents, error = fetch_file(url, 'badmd5')
    assert bool(error), 'should have errored'
    assert not contents

    contents, error = fetch_file('http://badhostname.willowgarage.com', 'md5sum')
    assert not contents
    assert bool(error), 'should have errored'


def test_download_rdmanifest():
    test_dir = get_test_dir()
    with open(os.path.join(test_dir, 'rep112-example.rdmanifest')) as f:
        expected = yaml.safe_load(f)

    from rosdep2.platforms.source import download_rdmanifest, DownloadFailed
    url = path_to_url(os.path.join(test_dir, 'rep112-example.rdmanifest'))
    contents, download_url = download_rdmanifest(url, REP112_MD5SUM)
    assert contents == expected
    assert download_url == url

    # test alt_url
    contents, download_url = download_rdmanifest('http://badhostname.willowgarage.com/', REP112_MD5SUM, alt_url=url)
    assert contents == expected
    assert download_url == url

    # test md5sum validate
    try:
        contents, error = download_rdmanifest(url, 'badmd5')
        assert False, 'should have errored'
    except DownloadFailed:
        pass

    # test download verify
    try:
        contents, error = download_rdmanifest('http://badhostname.willowgarage.com', 'fakemd5')
        assert False, 'should have errored'
    except DownloadFailed:
        pass


@pytest.mark.online
def test_install_from_file():
    from rosdep2.platforms.source import install_from_file
    f = os.path.join(get_test_dir(), 'noop-not-installed.rdmanifest')
    install_from_file(f)


@pytest.mark.online
def test_install_source():
    from rosdep2.platforms.source import install_source, SourceInstall
    resolved = SourceInstall()
    resolved.tarball = 'https://github.com/ros-infrastructure/rosdep/raw/master/test/source/foo.tar.gz'
    resolved.tarball_md5sum = 'fd34dc39f8f192b97fcc191fe0a6befc'
    resolved.install_command = """#!/bin/sh
exit 0
"""
    resolved.exec_path = ''
    install_source(resolved)



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Ken Conley/kwc@willowgarage.com

"""
Library for loading rosdep files from the ROS package/stack
filesystem.
"""

import os

import catkin_pkg.package
import rospkg

from .catkin_packages import VALID_DEPENDENCY_TYPES
from .loader import RosdepLoader

# Default view key is the view that packages that are not in stacks
# see. It is the root of all dependencies.  It is superceded by an
# explicit underlay_key.
DEFAULT_VIEW_KEY = '*default*'

# Implementation details: this API was originally conceived under the
# rosdep 1 design.  It has since been retrofitted for the rosdep 2
# design, which means it is a bit overbuilt.  There really is no need
# for a notion of views for rospkg -- all rospkgs have the same view.
# It we be nice to refactor this API into something much, much
# simpler, which would probably involve merging RosPkgLoader and
# SourcesListLoader.  RosPkgLoader would provide identification of
# resources and SourcesListLoader would build a *single* view that was
# no longer resource-dependent.


class RosPkgLoader(RosdepLoader):

    def __init__(self, rospack=None, rosstack=None, underlay_key=None, dependency_types=[]):
        """
        :param underlay_key: If set, all views loaded by this loader
            will depend on this key.
        """
        if rospack is None:
            rospack = rospkg.RosPack()
        if rosstack is None:
            rosstack = rospkg.RosStack()

        self._rospack = rospack
        self._rosstack = rosstack
        self._rosdep_yaml_cache = {}
        self._underlay_key = underlay_key

        # cache computed list of loadable resources
        self._loadable_resource_cache = None
        self._catkin_packages_cache = None

        default_dep_types = VALID_DEPENDENCY_TYPES - {'doc'}
        self.include_dep_types = VALID_DEPENDENCY_TYPES.intersection(set(dependency_types)) if dependency_types else default_dep_types

    def load_view(self, view_name, rosdep_db, verbose=False):
        """
        Load view data into *rosdep_db*. If the view has already
        been loaded into *rosdep_db*, this method does nothing.  If
        view has no rosdep data, it will be initialized with an empty
        data map.

        :raises: :exc:`InvalidData` if view rosdep.yaml is invalid
        :raises: :exc:`rospkg.ResourceNotFound` if view cannot be located

        :returns: ``True`` if view was loaded.  ``False`` if view
          was already loaded.
        """
        if rosdep_db.is_loaded(view_name):
            return
        if view_name not in self.get_loadable_views():
            raise rospkg.ResourceNotFound(view_name)
        elif view_name == 'invalid':
            raise rospkg.ResourceNotFound('FOUND' + view_name + str(self.get_loadable_views()))
        if verbose:
            print('loading view [%s] with rospkg loader' % (view_name))
        # chain into underlay if set
        if self._underlay_key:
            view_dependencies = [self._underlay_key]
        else:
            view_dependencies = []
        # no rospkg view has actual data
        rosdep_db.set_view_data(view_name, {}, view_dependencies, '<nodata>')

    def get_loadable_views(self):
        """
        'Views' map to ROS stack names.
        """
        return list(self._rosstack.list()) + [DEFAULT_VIEW_KEY]

    def get_loadable_resources(self):
        """
        'Resources' map to ROS packages names.
        """
        if not self._loadable_resource_cache:
            self._loadable_resource_cache = list(self._rospack.list())
        return self._loadable_resource_cache

    def get_catkin_paths(self):
        if not self._catkin_packages_cache:
            def find_catkin_paths(src):
                return [(x, src.get_path(x)) for x in
                        filter(lambda x: src.get_manifest(x).is_catkin, src.list())]
            self._catkin_packages_cache = dict(find_catkin_paths(self._rospack))
            self._catkin_packages_cache.update(find_catkin_paths(self._rosstack))
        return self._catkin_packages_cache

    def get_rosdeps(self, resource_name, implicit=True):
        """
        If *resource_name* is a stack, returns an empty list.

        :raises: :exc:`rospkg.ResourceNotFound` if *resource_name* cannot be found.
        """

        if resource_name in self.get_catkin_paths():
            pkg = catkin_pkg.package.parse_package(self.get_catkin_paths()[resource_name])
            pkg.evaluate_conditions(os.environ)
            deps = sum((getattr(pkg, '{}_depends'.format(d)) for d in self.include_dep_types), [])
            return [d.name for d in deps if d.evaluated_condition]
        elif resource_name in self.get_loadable_resources():
            rosdeps = set(self._rospack.get_rosdeps(resource_name, implicit=False))
            if implicit:
                # This resource is a manifest.xml, but it might depend on things with a package.xml
                # Make sure they get a chance to evaluate conditions
                for dep in self._rospack.get_depends(resource_name):
                    rosdeps = rosdeps.union(set(self.get_rosdeps(dep, implicit=True)))
            return list(rosdeps)
        elif resource_name in self._rosstack.list():
            # stacks currently do not have rosdeps of their own, implicit or otherwise
            return []
        else:
            raise rospkg.ResourceNotFound(resource_name)

    def is_metapackage(self, resource_name):
        if resource_name in self._rosstack.list():
            m = self._rosstack.get_manifest(resource_name)
            return m.is_catkin
        return False

    def get_view_key(self, resource_name):
        """
        Map *resource_name* to a view key.  In rospkg, this maps the
        DEFAULT_VIEW_KEY if *resource_name* exists.

        :raises: :exc:`rospkg.ResourceNotFound`
        """
        if (
            resource_name in self.get_catkin_paths() or
            resource_name in self.get_loadable_resources()
        ):
            return DEFAULT_VIEW_KEY
        else:
            raise rospkg.ResourceNotFound(resource_name)



# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Tully Foote/tfoote@willowgarage.com, Ken Conley/kwc@willowgarage.com

import os
import subprocess
import traceback

from rospkg.os_detect import OsDetect

from .core import rd_debug, RosdepInternalError, InstallFailed, print_bold, InvalidData

# kwc: InstallerContext is basically just a bunch of dictionaries with
# defined lookup methods.  It really encompasses two facets of a
# rosdep configuration: the pluggable nature of installers and
# platforms, as well as the resolution of the operating system for a
# specific machine.  It is possible to decouple those two notions,
# though there are some touch points over how this interfaces with the
# rospkg.os_detect library, i.e. how platforms can tweak these
# detectors and how the higher-level APIs can override them.


class InstallerContext(object):
    """
    :class:`InstallerContext` manages the context of execution for rosdep as it
    relates to the installers, OS detectors, and other extensible
    APIs.
    """

    def __init__(self, os_detect=None):
        """
        :param os_detect: (optional)
        :class:`rospkg.os_detect.OsDetect` instance to use for
          detecting platforms.  If `None`, default instance will be
          used.
        """
        # platform configuration
        self.installers = {}
        self.os_installers = {}
        self.default_os_installer = {}

        # stores configuration of which value to use for the OS version key (version number or codename)
        self.os_version_type = {}

        # OS detection and override
        if os_detect is None:
            os_detect = OsDetect()
        self.os_detect = os_detect
        self.os_override = None

        self.verbose = False

    def set_verbose(self, verbose):
        self.verbose = verbose

    def set_os_override(self, os_name, os_version):
        """
        Override the OS detector with *os_name* and *os_version*.  See
        :meth:`InstallerContext.detect_os`.

        :param os_name: OS name value to use, ``str``
        :param os_version: OS version value to use, ``str``
        """
        if self.verbose:
            print('overriding OS to [%s:%s]' % (os_name, os_version))
        self.os_override = os_name, os_version

    def get_os_version_type(self, os_name):
        return self.os_version_type.get(os_name, OsDetect.get_version)

    def set_os_version_type(self, os_name, version_type):
        if not hasattr(version_type, '__call__'):
            raise ValueError('version type should be a method')
        self.os_version_type[os_name] = version_type

    def get_os_name_and_version(self):
        """
        Get the OS name and version key to use for resolution and
        installation.  This will be the detected OS name/version
        unless :meth:`InstallerContext.set_os_override()` has been
        called.

        :returns: (os_name, os_version), ``(str, str)``
        """
        if self.os_override:
            return self.os_override
        else:
            os_name = self.os_detect.get_name()
            os_key = self.get_os_version_type(os_name)
            os_version = os_key(self.os_detect)
            return os_name, os_version

    def get_os_detect(self):
        """
        :returns os_detect: :class:`OsDetect` instance used for
          detecting platforms.
        """
        return self.os_detect

    def set_installer(self, installer_key, installer):
        """
        Set the installer to use for *installer_key*.  This will
        replace any existing installer associated with the key.
        *installer_key* should be the same key used for the
        ``rosdep.yaml`` package manager key.  If *installer* is
        ``None``, this will delete any existing associated installer
        from this context.

        :param installer_key: key/name to associate with installer, ``str``
        :param installer: :class:`Installer` implementation, ``class``.
        :raises: :exc:`TypeError` if *installer* is not a subclass of
          :class:`Installer`
        """
        if installer is None:
            del self.installers[installer_key]
            return
        if not isinstance(installer, Installer):
            raise TypeError('installer must be a instance of Installer')
        if self.verbose:
            print('registering installer [%s]' % (installer_key))
        self.installers[installer_key] = installer

    def get_installer(self, installer_key):
        """
        :returns: :class:`Installer` class associated with *installer_key*.
        :raises: :exc:`KeyError` If not associated installer
        :raises: :exc:`InstallFailed` If installer cannot produce an install command (e.g. if installer is not installed)
        """
        return self.installers[installer_key]

    def get_installer_keys(self):
        """
        :returns: list of registered installer keys
        """
        return self.installers.keys()

    def get_os_keys(self):
        """
        :returns: list of OS keys that have registered with this context, ``[str]``
        """
        return self.os_installers.keys()

    def add_os_installer_key(self, os_key, installer_key):
        """
        Register an installer for the specified OS.  This will fail
        with a :exc:`KeyError` if no :class:`Installer` can be found
        with the associated *installer_key*.

        :param os_key: Key for OS
        :param installer_key: Key for installer to add to OS
        :raises: :exc:`KeyError`: if installer for *installer_key*
          is not set.
        """
        # validate, will throw KeyError
        self.get_installer(installer_key)
        if self.verbose:
            print('add installer [%s] to OS [%s]' % (installer_key, os_key))
        if os_key in self.os_installers:
            self.os_installers[os_key].append(installer_key)
        else:
            self.os_installers[os_key] = [installer_key]

    def get_os_installer_keys(self, os_key):
        """
        Get list of installer keys registered for the specified OS.
        These keys can be resolved by calling
        :meth:`InstallerContext.get_installer`.

        :param os_key: Key for OS
        :raises: :exc:`KeyError`: if no information for OS *os_key* is registered.
        """
        if os_key in self.os_installers:
            return self.os_installers[os_key][:]
        else:
            raise KeyError(os_key)

    def set_default_os_installer_key(self, os_key, installer_key):
        """
        Set the default OS installer to use for OS.
        :meth:`InstallerContext.add_os_installer` must have previously
        been called with the same arguments.

        :param os_key: Key for OS
        :param installer_key: Key for installer to add to OS
        :raises: :exc:`KeyError`: if installer for *installer_key*
          is not set or if OS for *os_key* has no associated installers.
        """
        if os_key not in self.os_installers:
            raise KeyError('unknown OS: %s' % (os_key))
        if not hasattr(installer_key, '__call__'):
            raise ValueError('version type should be a method')
        if not installer_key(self.os_detect) in self.os_installers[os_key]:
            raise KeyError('installer [%s] is not associated with OS [%s]. call add_os_installer_key() first' % (installer_key(self.os_detect), os_key))
        if self.verbose:
            print('set default installer [%s] for OS [%s]' % (installer_key(self.os_detect), os_key,))
        self.default_os_installer[os_key] = installer_key

    def get_default_os_installer_key(self, os_key):
        """
        Get the default OS installer key to use for OS, or ``None`` if
        there is no default.

        :param os_key: Key for OS
        :returns: :class:`Installer`
        :raises: :exc:`KeyError`: if no information for OS *os_key* is registered.
        """
        if os_key not in self.os_installers:
            raise KeyError('unknown OS: %s' % (os_key))
        try:
            installer_key = self.default_os_installer[os_key](self.os_detect)
            if installer_key not in self.os_installers[os_key]:
                raise KeyError('installer [%s] is not associated with OS [%s]. call add_os_installer_key() first' % (installer_key, os_key))
            # validate, will throw KeyError
            self.get_installer(installer_key)
            return installer_key
        except KeyError:
            return None


class Installer(object):
    """
    The :class:`Installer` API is designed around opaque *resolved*
    parameters. These parameters can be any type of sequence object,
    but they must obey set arithmetic.  They should also implement
    ``__str__()`` methods so they can be pretty printed.
    """

    def is_installed(self, resolved_item):
        """
        :param resolved: resolved installation item. NOTE: this is a single item,
          not a list of items like the other APIs, ``opaque``.
        :returns: ``True`` if all of the *resolved* items are installed on
          the local system
        """
        raise NotImplementedError('is_installed', resolved_item)

    def get_install_command(self, resolved, interactive=True, reinstall=False, quiet=False, oneshot=[]):
        """
        :param resolved: list of resolved installation items, ``[opaque]``
        :param interactive: If `False`, disable interactive prompts,
          e.g. Pass through ``-y`` or equivalant to package manager.
        :param reinstall: If `True`, install everything even if already installed
        """
        raise NotImplementedError('get_package_install_command', resolved, interactive, reinstall, quiet, oneshot)

    def get_depends(self, rosdep_args):
        """
        :returns: list of dependencies on other rosdep keys.  Only
          necessary if the package manager doesn't handle
          dependencies.
        """
        return []  # Default return empty list

    def resolve(self, rosdep_args_dict):
        """
        :param rosdep_args_dict: argument dictionary to the rosdep rule for this package manager
        :returns: [resolutions].  resolved objects should be printable to a user, but are otherwise opaque.
        """
        raise NotImplementedError('Base class resolve', rosdep_args_dict)

    def unique(self, *resolved_rules):
        """
        Combine the resolved rules into a unique list.  This
        is meant to combine the results of multiple calls to
        :meth:`PackageManagerInstaller.resolve`.

        Example::

            resolved1 = installer.resolve(args1)
            resolved2 = installer.resolve(args2)
            resolved = installer.unique(resolved1, resolved2)

        :param resolved_rules: resolved arguments.  Resolved
          arguments must all be from this :class:`Installer` instance.
        """
        raise NotImplementedError('Base class unique', resolved_rules)


class PackageManagerInstaller(Installer):
    """
    General form of a package manager :class:`Installer`
    implementation that assumes:

     - installer rosdep args spec is a list of package names stored with the key "packages"
     - a detect function exists that can return a list of packages that are installed

    Also, if *supports_depends* is set to ``True``:

     - installer rosdep args spec can also include dependency specification with the key "depends"
    """

    def __init__(self, detect_fn, supports_depends=False):
        """
        :param supports_depends: package manager supports dependency key
        :param detect_fn: function that for a given list of packages determines
        the list of installed packages.
        """
        self.detect_fn = detect_fn
        self.supports_depends = supports_depends
        self.as_root = True
        self.sudo_command = 'sudo -H' if hasattr(os, 'geteuid') and os.geteuid() != 0 else ''

    def elevate_priv(self, cmd):
        """
        Prepend *self.sudo_command* to the command if *self.as_root* is ``True``.

        :param list cmd: list of strings comprising the command
        :returns: a list of commands
        """
        return (self.sudo_command.split() if self.as_root else []) + cmd

    def resolve(self, rosdep_args):
        """
        See :meth:`Installer.resolve()`
        """
        packages = None
        if type(rosdep_args) is dict:
            packages = rosdep_args.get('packages', [])
            if isinstance(packages, str):
                packages = packages.split()
        elif isinstance(rosdep_args, str):
            packages = rosdep_args.split(' ')
        elif type(rosdep_args) is list:
            packages = rosdep_args
        else:
            raise InvalidData('Invalid rosdep args: %s' % (rosdep_args))
        return packages

    def unique(self, *resolved_rules):
        """
        See :meth:`Installer.unique()`
        """
        s = set()
        for resolved in resolved_rules:
            s.update(resolved)
        return sorted(s)

    def get_packages_to_install(self, resolved, reinstall=False):
        """
        Return a list of packages (out of *resolved*) that still need to get
        installed.
        """
        if reinstall:
            return resolved
        if not resolved:
            return []
        else:
            detected = self.detect_fn(resolved)
            return [x for x in resolved if x not in detected]

    def is_installed(self, resolved_item):
        """
        Check if a given package was installed.
        """
        return not self.get_packages_to_install([resolved_item])

    def get_version_strings(self):
        """
        Return a list of version information strings.

        Where each string is of the form "<installer> <version string>".
        For example, ["apt-get x.y.z"] or ["pip x.y.z", "setuptools x.y.z"].
        """
        raise NotImplementedError('subclasses must implement get_version_strings method')

    def get_install_command(self, resolved, interactive=True, reinstall=False, quiet=False, oneshot=[]):
        raise NotImplementedError('subclasses must implement', resolved, interactive, reinstall, quiet, oneshot)

    def get_depends(self, rosdep_args):
        """
        :returns: list of dependencies on other rosdep keys.  Only
          necessary if the package manager doesn't handle
          dependencies.
        """
        if self.supports_depends and type(rosdep_args) is dict:
            return rosdep_args.get('depends', [])
        return []  # Default return empty list


def normalize_uninstalled_to_list(uninstalled):
    uninstalled_dependencies = []
    for pkg_or_list in [v for k, v in uninstalled.items()]:
        if isinstance(pkg_or_list, list):
            for pkg in pkg_or_list:
                uninstalled_dependencies.append(str(pkg))
        else:
            uninstalled_dependencies.append(str(pkg))
    return uninstalled_dependencies


class RosdepInstaller(object):

    def __init__(self, installer_context):
        self.installer_context = installer_context

    def get_uninstalled(self, resolutions, verbose=False):
        """
        Get list of system dependencies that have not been installed
        as well as a list of errors from performing the resolution.
        This is a bulk API in order to provide performance
        optimizations in checking install state.

        :param resolutions: ``[(str, [str])]``.  List of resolution tuples.
          A resolution tuple's first element is the installer key (e.g.: apt or homebrew)
          and the second element is a list of installer specific resolution values.
        :param implicit: Install implicit (recursive) dependencies of
            resources.  Default ``False``.

        :returns: (uninstalled, errors), ``({str: [opaque]}, {str: ResolutionError})``.
          Uninstalled is a dictionary with the installer_key as the key.
        :raises: :exc:`RosdepInternalError`
        """

        # for each installer, figure out what is left to install
        uninstalled = {}
        for installer_key, resolved in resolutions.items():  # py3k
            if verbose:
                print('resolution: %s [%s]' % (installer_key, ', '.join([str(r) for r in resolved])))
            try:
                installer = self.installer_context.get_installer(installer_key)
            except KeyError as e:  # lookup has to be buggy to cause this
                raise RosdepInternalError(e)
            try:
                packages_to_install = installer.get_packages_to_install(resolved)
            except Exception as e:
                rd_debug(traceback.format_exc())
                raise RosdepInternalError(e, message='Bad installer [%s]: %s' % (installer_key, e))

            # only create key if there is something to do
            if packages_to_install:
                if installer_key not in uninstalled:
                    uninstalled[installer_key] = set()
                uninstalled[installer_key].update(packages_to_install)
            if verbose:
                print('uninstalled: [%s]' % (', '.join([str(p) for p in packages_to_install])))

        return uninstalled

    def install(self, uninstalled, interactive=True, simulate=False, oneshot=[],
                continue_on_error=False, reinstall=False, verbose=False, quiet=False):
        """
        Install the uninstalled rosdeps.  This API is for the bulk
        workflow of rosdep (see example below).  For a more targeted
        install API, see :meth:`RosdepInstaller.install_resolved`.

        :param uninstalled: uninstalled value from
          :meth:`RosdepInstaller.get_uninstalled`.  Value is a
          dictionary mapping installer key to a dictionary with resolution
          data, ``{str: {str: vals}}``
        :param interactive: If ``False``, suppress
          interactive prompts (e.g. by passing '-y' to ``apt``).
        :param simulate: If ``False`` simulate installation
          without actually executing.
        :param continue_on_error: If ``True``, continue installation
          even if an install fails.  Otherwise, stop after first
          installation failure.
        :param reinstall: If ``True``, install dependencies if even
          already installed (default ``False``).

        :raises: :exc:`InstallFailed` if any rosdeps fail to install
          and *continue_on_error* is ``False``.
        :raises: :exc:`KeyError` If *uninstalled* value has invalid
          installer keys

        Example::

            uninstalled, errors = installer.get_uninstalled(packages)
            installer.install(uninstalled)
        """
        if verbose:
            print(
                'install options: reinstall[%s] simulate[%s] interactive[%s]' %
                (reinstall, simulate, interactive)
            )
            uninstalled_list = normalize_uninstalled_to_list(uninstalled)
            print('install: uninstalled keys are %s' % ', '.join(uninstalled_list))

        failures = []
        for installer_key, resolved in uninstalled.items():
            try:
                self.install_resolved(installer_key, resolved, simulate=simulate,
                                      interactive=interactive, reinstall=reinstall, continue_on_error=continue_on_error,
                                      verbose=verbose, quiet=quiet, oneshot=oneshot)
            except InstallFailed as e:
                if not continue_on_error:
                    raise
                else:
                    # accumulate errors
                    failures.extend(e.failures)
        if failures:
            raise InstallFailed(failures=failures)

    def install_resolved(self, installer_key, resolved, simulate=False, interactive=True, oneshot=[],
                         reinstall=False, continue_on_error=False, verbose=False, quiet=False):
        """
        Lower-level API for installing a rosdep dependency.  The
        rosdep keys have already been resolved to *installer_key* and
        *resolved* via :exc:`RosdepLookup` or other means.

        :param installer_key: Key for installer to apply to *resolved*, ``str``
        :param resolved: Opaque resolution list from :class:`RosdepLookup`.
        :param interactive: If ``True``, allow interactive prompts (default ``True``)
        :param simulate: If ``True``, don't execute installation commands, just print to screen.
        :param reinstall: If ``True``, install dependencies if even
          already installed (default ``False``).
        :param verbose: If ``True``, print verbose output to screen (default ``False``)
        :param quiet: If ``True``, supress output except for errors (default ``False``)

        :raises: :exc:`InstallFailed` if any of *resolved* fail to install.
        """
        installer_context = self.installer_context
        installer = installer_context.get_installer(installer_key)
        command = installer.get_install_command(resolved, interactive=interactive, reinstall=reinstall, quiet=quiet, oneshot=oneshot)
        if not command:
            if verbose:
                print('#No packages to install')
            return

        if simulate:
            print('#[%s] Installation commands:' % (installer_key))
            for sub_command in command:
                if isinstance(sub_command[0], list):
                    sub_cmd_len = len(sub_command)
                    for i, cmd in enumerate(sub_command):
                        print("  '%s' (alternative %d/%d)" % (' '.join(cmd), i + 1, sub_cmd_len))
                else:
                    print('  ' + ' '.join(sub_command))

        # nothing left to do for simulation
        if simulate:
            return

        def run_command(command, installer_key, failures, verbose):
            # always echo commands to screen
            print_bold('executing command [%s]' % ' '.join(command))
            result = subprocess.call(command)
            if verbose:
                print('command return code [%s]: %s' % (' '.join(command), result))
            if result != 0:
                failures.append((installer_key, 'command [%s] failed' % (' '.join(command))))
            return result

        # run each install command set and collect errors
        failures = []
        for sub_command in command:
            if isinstance(sub_command[0], list):  # list of alternatives
                alt_failures = []
                for alt_command in sub_command:
                    result = run_command(alt_command, installer_key, alt_failures, verbose)
                    if result == 0:  # one successsfull command is sufficient
                        alt_failures = []  # clear failuers from other alternatives
                        break
                failures.extend(alt_failures)
            else:
                result = run_command(sub_command, installer_key, failures, verbose)
            if result != 0:
                if not continue_on_error:
                    raise InstallFailed(failures=failures)

        # test installation of each
        for r in resolved:
            if not installer.is_installed(r):
                failures.append((installer_key, 'Failed to detect successful installation of [%s]' % (r)))
        # finalize result
        if failures:
            raise InstallFailed(failures=failures)
        elif verbose:
            print('#successfully installed')



# Copyright (c) 2009, Willow Garage, Inc.
# Copyright (c) 2019, Kei Okada
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import subprocess

from ..core import InstallFailed
from ..installers import PackageManagerInstaller
from ..shell_utils import read_stdout

# npm package manager key
NPM_INSTALLER = 'npm'


def register_installers(context):
    context.set_installer(NPM_INSTALLER, NpmInstaller())


def is_npm_installed():
    try:
        subprocess.Popen(['npm'], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
        return True
    except OSError:
        return False


class NpmInstaller(PackageManagerInstaller):
    """
    :class:`Installer` support for npm.
    """

    def __init__(self):
        super(NpmInstaller, self).__init__(self.npm_detect, supports_depends=True)

    def npm_detect(self, pkgs, exec_fn=None):
        """
        Given a list of package, return the list of installed packages.

        :param exec_fn: function to execute Popen and read stdout (for testing)
        """
        if exec_fn is None:
            exec_fn = read_stdout

        # npm list -parseable returns [dir, dir/node_modules/path, dir/node_modules/path, ...]
        if self.as_root:
            cmd = ['npm', 'list', '-g']
        else:
            cmd = ['npm', 'list']
        pkg_list = exec_fn(cmd + ['-parseable']).split('\n')

        ret_list = []
        for pkg in pkg_list[1:]:
            pkg_row = pkg.split('/')
            if pkg_row[-1] in pkgs:
                ret_list.append(pkg_row[-1])
        return ret_list

    def get_version_strings(self):
        npm_version = subprocess.check_output(['npm', '--version']).strip().decode()
        return ['npm {}'.format(npm_version)]

    def get_install_command(self, resolved, interactive=True, reinstall=False, quiet=False):
        if not is_npm_installed():
            raise InstallFailed((NPM_INSTALLER, 'npm is not installed'))
        packages = self.get_packages_to_install(resolved, reinstall=reinstall)
        if not packages:
            return []
        if self.as_root:
            cmd = ['npm', 'install', '-g']
        else:
            cmd = ['npm', 'install']

        return [self.elevate_priv(cmd + [p]) for p in packages]



# Copyright (c) 2018, SEQSENSE, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Atsushi Watanabe/atsushi.w@ieee.org

import os

from rospkg.os_detect import OS_ALPINE

from .pip import PIP_INSTALLER
from .source import SOURCE_INSTALLER
from ..installers import PackageManagerInstaller
from ..shell_utils import read_stdout

APK_INSTALLER = 'apk'


def register_installers(context):
    context.set_installer(APK_INSTALLER, ApkInstaller())


def register_platforms(context):
    context.add_os_installer_key(OS_ALPINE, APK_INSTALLER)
    context.add_os_installer_key(OS_ALPINE, PIP_INSTALLER)
    context.add_os_installer_key(OS_ALPINE, SOURCE_INSTALLER)
    context.set_default_os_installer_key(OS_ALPINE, lambda self: APK_INSTALLER)
    context.set_os_version_type(OS_ALPINE, lambda self: '.'.join(self.get_version().split('.')[:2]))


def apk_detect(pkgs, exec_fn=read_stdout):
    """
    Given a list of packages, return a list of which are already installed.

    :param exec_fn: function to execute Popen and read stdout (for testing)
    """

    if not pkgs:
        return []

    # Get installed package names:
    cmd_installed = ['apk', 'info', '--installed']
    cmd_installed.extend(pkgs)
    origin_packages = exec_fn(cmd_installed).splitlines()

    # Resolve alias names of replaced packages.
    cmd_replaces = ['apk', 'info', '--installed', '--replaces']
    cmd_replaces.extend(pkgs)
    # This command will respond like:
    #
    #    $ apk info --installed --replaces boost-atomic boost-filesystem
    #    boost1.76-atomic-1.76.0-r0 replaces:
    #    boost-atomic
    #
    #    boost1.76-filesystem-1.76.0-r0 replaces:
    #    boost-filesystem
    #
    #    $
    replaced_packages = []
    reading = False
    for line in exec_fn(cmd_replaces).splitlines():
        if line.endswith(' replaces:'):
            reading = True
        elif line == '':
            reading = False
        elif reading:
            replaced_packages.append(line.strip())

    return origin_packages + replaced_packages


class ApkInstaller(PackageManagerInstaller):

    def __init__(self):
        super(ApkInstaller, self).__init__(apk_detect)

    def get_install_command(self, resolved, interactive=True, reinstall=False, quiet=False):
        pkgs = self.get_packages_to_install(resolved, reinstall=reinstall)
        if not pkgs:
            return []

        cmd = self.elevate_priv(['apk', 'add'])

        if interactive:
            cmd.append('--interactive')
        if quiet:
            cmd.append('--quiet')

        cmd.extend(pkgs)

        return [cmd]



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Ken Conley/kwc@willowgarage.com

import os
import yaml
from unittest.mock import Mock

from rospkg import RosPack, RosStack


def get_test_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'tree'))


def get_rospkg():
    # configure inside of the test tree
    test_dir = get_test_dir()
    ros_root = os.path.join(test_dir, 'ros')
    ros_package_path = os.path.join(test_dir, 'stacks')
    ros_paths = [ros_root, ros_package_path]
    rospack = RosPack(ros_paths=ros_paths)
    rosstack = RosStack(ros_paths=ros_paths)
    return rospack, rosstack


def test_RosPkgLoader():
    from rospkg import ResourceNotFound

    from rosdep2.model import RosdepDatabase
    from rosdep2.rospkg_loader import RosPkgLoader, DEFAULT_VIEW_KEY
    from rosdep2.loader import InvalidData

    # Due to rosdep 2/REP 125 changes, this test is a bit overbuilt.
    # All stacks return the same.

    # tripwire
    loader = RosPkgLoader()
    assert loader._rospack is not None
    assert loader._rosstack is not None

    # configure inside of the test tree
    rospack, rosstack = get_rospkg()
    ros_root = rosstack.get_path('ros')
    loader = RosPkgLoader(rospack, rosstack)
    assert loader._rospack == rospack
    assert loader._rosstack == rosstack

    # test with mock db
    rosdep_db = Mock(spec=RosdepDatabase)
    rosdep_db.is_loaded.return_value = False

    # test with no rosdep.yaml stack
    loader.load_view('empty', rosdep_db)
    rosdep_db.is_loaded.assert_called_with('empty')
    rosdep_db.set_view_data.assert_called_with('empty', {}, [], '<nodata>')

    # test invalid stack
    try:
        loader.load_view('does not exist', rosdep_db)
        assert False, 'should have raised'
    except ResourceNotFound as e:
        pass

    # Test with default view key
    loader.load_view(DEFAULT_VIEW_KEY, rosdep_db)
    rosdep_db.set_view_data.assert_called_with(DEFAULT_VIEW_KEY, {}, [], '<nodata>')

    # test with complicated ros stack.
    loader.load_view('ros', rosdep_db)
    rosdep_db.is_loaded.assert_called_with('ros')
    rosdep_db.set_view_data.assert_called_with('ros', {}, [], '<nodata>')

    # test call on db that is already loaded
    rosdep_db.reset_mock()
    rosdep_db.is_loaded.return_value = True
    loader.load_view('ros', rosdep_db)
    rosdep_db.is_loaded.assert_called_with('ros')
    assert rosdep_db.set_view_data.call_args_list == []

    # test get_view_key, always the same return value
    assert loader.get_view_key('stack1_p1') == DEFAULT_VIEW_KEY
    assert loader.get_view_key('stackless') == DEFAULT_VIEW_KEY
    try:
        loader.get_view_key('fake')
        assert False, 'should error'
    except ResourceNotFound:
        pass


def test_RosPkgLoader_with_underlay_key():
    from rospkg import ResourceNotFound

    from rosdep2.model import RosdepDatabase
    from rosdep2.rospkg_loader import RosPkgLoader, DEFAULT_VIEW_KEY
    from rosdep2.loader import InvalidData

    # configure inside of the test tree
    rospack, rosstack = get_rospkg()
    ros_root = rosstack.get_path('ros')
    loader = RosPkgLoader(rospack, rosstack, underlay_key='underlay-key')
    assert loader._rospack == rospack
    assert loader._rosstack == rosstack

    # test with mock db
    rosdep_db = Mock(spec=RosdepDatabase)
    rosdep_db.is_loaded.return_value = False

    # test with no rosdep.yaml stack
    loader.load_view('empty', rosdep_db)
    rosdep_db.is_loaded.assert_called_with('empty')
    rosdep_db.set_view_data.assert_called_with('empty', {}, ['underlay-key'], '<nodata>')

    # test invalid stack
    try:
        loader.load_view('does not exist', rosdep_db)
        assert False, 'should have raised'
    except ResourceNotFound as e:
        pass

    # test with complicated ros stack.
    loader.load_view('ros', rosdep_db)
    rosdep_db.is_loaded.assert_called_with('ros')
    rosdep_db.set_view_data.assert_called_with('ros', {}, ['underlay-key'], '<nodata>')

    # test call on db that is already loaded
    rosdep_db.reset_mock()
    rosdep_db.is_loaded.return_value = True
    loader.load_view('ros', rosdep_db)
    rosdep_db.is_loaded.assert_called_with('ros')
    assert rosdep_db.set_view_data.call_args_list == []

    # test get_view_key
    assert loader.get_view_key('stack1_p1') == DEFAULT_VIEW_KEY
    assert loader.get_view_key('stackless') == DEFAULT_VIEW_KEY
    try:
        loader.get_view_key('fake')
        assert False, 'should error'
    except ResourceNotFound:
        pass


def test_RosPkgLoader_get_loadable():
    from rosdep2.rospkg_loader import RosPkgLoader

    rospack, rosstack = get_rospkg()
    loader = RosPkgLoader(rospack, rosstack)
    assert loader._rospack == rospack
    assert loader._rosstack == rosstack

    keys = loader.get_loadable_resources()
    for p in ['stack1_p1', 'stack1_p2', 'stack1_p3']:
        assert p in keys
    keys = loader.get_loadable_views()
    for s in ['ros', 'empty', 'invalid', 'stack1']:
        assert s in keys



# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from contextlib import contextmanager
import os
import sys
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

import rospkg

import unittest
from unittest.mock import DEFAULT, patch

from rosdep2 import create_default_installer_context
from rosdep2 import main
from rosdep2.ament_packages import AMENT_PREFIX_PATH_ENV_VAR
from rosdep2.main import rosdep_main
from rosdep2.main import setup_proxy_opener


GITHUB_BASE_URL = 'https://github.com/ros/rosdistro/raw/master/rosdep/base.yaml'
GITHUB_PYTHON_URL = 'https://github.com/ros/rosdistro/raw/master/rosdep/python.yaml'


def get_test_dir():
    return os.path.abspath(os.path.dirname(__file__))


def get_test_tree_dir():
    return os.path.abspath(os.path.join(get_test_dir(), 'tree'))


def get_test_catkin_tree_dir():
    return os.path.abspath(os.path.join(get_test_tree_dir(), 'catkin'))


def get_cache_dir():
    p = os.path.join(get_test_dir(), 'sources_cache')
    assert os.path.isdir(p)
    return p


@contextmanager
def fakeout():
    realstdout = sys.stdout
    realstderr = sys.stderr
    fakestdout = StringIO()
    fakestderr = StringIO()
    sys.stdout = fakestdout
    sys.stderr = fakestderr
    yield fakestdout, fakestderr
    sys.stdout = realstdout
    sys.stderr = realstderr

# the goal of these tests is only to test that we are wired into the
# APIs.  More exhaustive tests are at the unit level.


class TestRosdepMain(unittest.TestCase):

    def setUp(self):
        if 'ROSDEP_DEBUG' in os.environ:
            del os.environ['ROSDEP_DEBUG']
        self.old_rr = rospkg.get_ros_root()
        self.old_rpp = rospkg.get_ros_package_path()
        self.old_app = os.getenv(AMENT_PREFIX_PATH_ENV_VAR, None)
        if 'ROS_ROOT' in os.environ:
            del os.environ['ROS_ROOT']
        os.environ['ROS_PACKAGE_PATH'] = os.path.join(get_test_tree_dir())
        os.environ[AMENT_PREFIX_PATH_ENV_VAR] = os.path.join(get_test_tree_dir(), 'ament')
        if 'ROS_PYTHON_VERSION' not in os.environ:
            # avoid `test_check` failure due to warning on stderr
            os.environ['ROS_PYTHON_VERSION'] = sys.version[0]

    def tearDown(self):
        if self.old_rr is not None:
            os.environ['ROS_ROOT'] = self.old_rr
        if self.old_rpp is not None:
            os.environ['ROS_PACKAGE_PATH'] = self.old_rpp
        if self.old_app is not None:
            os.environ[AMENT_PREFIX_PATH_ENV_VAR] = self.old_app

    def test_bad_commands(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]
        for commands in [[], ['fake', 'something'], ['check'], ['install', '-a', 'rospack_fake'],
                         ['check', 'rospack_fake', '--os', 'ubuntulucid'],
                         ]:
            try:
                rosdep_main(commands + cmd_extras)
                assert False, 'system exit should have occurred'
            except SystemExit:
                pass

    def test_check(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]

        with fakeout() as b:
            try:
                rosdep_main(['check', 'python_dep'] + cmd_extras)
            except SystemExit:
                assert False, 'system exit occurred: %s\n%s' % (b[0].getvalue(), b[1].getvalue())

            stdout, stderr = b
            assert stdout.getvalue().strip() == 'All system dependencies have been satisfied', stdout.getvalue()
        try:
            context = create_default_installer_context()
            override = '%s:%s' % context.get_os_name_and_version()
            with fakeout() as b:
                rosdep_main(['check', 'python_dep', '--os', override] + cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == 'All system dependencies have been satisfied'
        except SystemExit:
            assert False, 'system exit occurred'

        # this used to abort, but now rosdep assumes validity for even empty stack args
        try:
            with fakeout() as b:
                rosdep_main(['check', 'packageless'] + cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == 'All system dependencies have been satisfied'
        except SystemExit:
            assert False, 'system exit occurred'

        try:
            rosdep_main(['check', 'nonexistent'] + cmd_extras)
            assert False, 'system exit should have occurred'
        except SystemExit:
            pass

    @patch('rosdep2.platforms.debian.read_stdout')
    @patch('rosdep2.installers.os.geteuid', return_value=1)
    def test_install(self, mock_geteuid, mock_read_stdout):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]
        catkin_tree = get_test_catkin_tree_dir()

        def read_stdout(cmd, capture_stderr=False):
            if cmd[0] == 'apt-cache' and cmd[1] == 'showpkg':
                result = ''
            elif cmd[0] == 'dpkg-query':
                if cmd[-1] == 'python3-dev':
                    result = "'python3-dev install ok installed\n'"
                else:
                    result = '\n'.join(['dpkg-query: no packages found matching %s' % f for f in cmd[3:]])

            if capture_stderr:
                return result, ''
            return result

        try:
            mock_read_stdout.side_effect = read_stdout
            # python must have already been installed
            with fakeout() as b:
                rosdep_main(['install', 'python_dep'] + cmd_extras)
                stdout, stderr = b
                assert 'All required rosdeps installed' in stdout.getvalue(), stdout.getvalue()
            with fakeout() as b:
                rosdep_main(['install', 'python_dep', '-r'] + cmd_extras)
                stdout, stderr = b
                assert 'All required rosdeps installed' in stdout.getvalue(), stdout.getvalue()
            # with fakeout() as b:
            #     rosdep_main([
            #         'install', '-s', '-i',
            #         '--os', 'ubuntu:lucid',
            #         '--rosdistro', 'fuerte',
            #         '--from-paths', catkin_tree
            #     ] + cmd_extras)
            #     stdout, stderr = b
            #     expected = [
            #         '#[apt] Installation commands:',
            #         '  sudo -H apt-get install ros-fuerte-catkin',
            #         '  sudo -H apt-get install libboost1.40-all-dev',
            #         '  sudo -H apt-get install libeigen3-dev',
            #         '  sudo -H apt-get install libtinyxml-dev',
            #         '  sudo -H apt-get install libltdl-dev',
            #         '  sudo -H apt-get install libtool',
            #         '  sudo -H apt-get install libcurl4-openssl-dev',
            #     ]
            #     lines = stdout.getvalue().splitlines()
            #     assert set(lines) == set(expected), lines
            with fakeout() as b:
                rosdep_main([
                    'install', '-s', '-i',
                    # '--one-shot', "apt", # oneshot is default now
                    '--os', 'ubuntu:lucid',
                    '--rosdistro', 'fuerte',
                    '--from-paths', catkin_tree
                ] + cmd_extras)
                stdout, stderr = b
                # the output is sorted in ascending order
                expected = [
                    '#[apt] Installation commands:',
                    '  sudo -H apt-get install libboost1.40-all-dev libcurl4-openssl-dev'
                    ' libeigen3-dev libltdl-dev libtinyxml-dev libtool ros-fuerte-catkin',
                ]
                lines = stdout.getvalue().splitlines()
                assert set(lines) == set(expected), lines
        except SystemExit:
            assert False, 'system exit occurred: ' + b[1].getvalue()
        try:
            rosdep_main(['install', 'nonexistent'])
            assert False, 'system exit should have occurred'
        except SystemExit:
            pass

    def test_where_defined(self):
        try:
            sources_cache = get_cache_dir()
            expected = GITHUB_PYTHON_URL
            for command in (['where_defined', 'testpython'], ['where_defined', 'testpython']):
                with fakeout() as b:
                    # set os to ubuntu so this test works on different platforms
                    rosdep_main(command + ['-c', sources_cache, '--os=ubuntu:lucid'])
                    stdout, stderr = b
                    output = stdout.getvalue().strip()
                    assert output == expected, output
        except SystemExit:
            assert False, 'system exit occurred'

    def test_what_needs(self):
        try:
            sources_cache = get_cache_dir()
            cmd_extras = ['-c', sources_cache]
            expected = ['python_dep']
            with fakeout() as b:
                rosdep_main(['what-needs', 'testpython'] + cmd_extras)
                stdout, stderr = b
                output = stdout.getvalue().strip()
                assert output.split('\n') == expected
            expected = ['python_dep']
            with fakeout() as b:
                rosdep_main(['what_needs', 'testpython', '--os', 'ubuntu:lucid', '--verbose'] + cmd_extras)
                stdout, stderr = b
                output = stdout.getvalue().strip()
                assert output.split('\n') == expected
        except SystemExit:
            assert False, 'system exit occurred'

    def test_keys(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]

        try:
            with fakeout() as b:
                rosdep_main(['keys', 'rospack_fake'] + cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == 'testtinyxml', stdout.getvalue()
                assert not stderr.getvalue(), stderr.getvalue()
            with fakeout() as b:
                rosdep_main(['keys', 'rospack_fake', '--os', 'ubuntu:lucid', '--verbose'] + cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == 'testtinyxml', stdout.getvalue()
            with fakeout() as b:
                rosdep_main(['keys', 'another_catkin_package'] + cmd_extras + ['-i'])
                stdout, stderr = b
                assert stdout.getvalue().strip() == 'catkin', stdout.getvalue()
            with fakeout() as b:
                rosdep_main(['keys', 'multi_dep_type_catkin_package', '-t', 'test', '-t', 'doc'] + cmd_extras)
                stdout, stderr = b
                output_keys = set(stdout.getvalue().split())
                expected_keys = set(['curl', 'epydoc'])
                assert output_keys == expected_keys, stdout.getvalue()
        except SystemExit:
            assert False, 'system exit occurred'
        try:
            rosdep_main(['keys', 'nonexistent'] + cmd_extras)
            assert False, 'system exit should have occurred'
        except SystemExit:
            pass

    def test_search(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]

        try:
            with fakeout() as b:
                rosdep_main(['search', 'curl', '--os=debian:squeeze'] + cmd_extras)
                stdout, stderr = b
                assert 'Closest keys' in stdout.getvalue(), stdout.getvalue()
                assert 'curl' in stdout.getvalue(), stdout.getvalue()
                assert 'Closest packages' not in stdout.getvalue(), stdout.getvalue()
                assert not stderr.getvalue(), stderr.getvalue()
            with fakeout() as b:
                rosdep_main(['search', 'libeigen3-dev', '--os=ubuntu:noble'] + cmd_extras)
                stdout, stderr = b
                assert 'Closest keys' not in stdout.getvalue(), stdout.getvalue()
                assert 'Closest packages' in stdout.getvalue(), stdout.getvalue()
                assert 'eigen:' in stdout.getvalue(), stdout.getvalue()
                assert not stderr.getvalue(), stderr.getvalue()
        except SystemExit:
            assert False, 'system exit occurred'
        try:
            rosdep_main(['search', 'libeigen3-dev', '--os=debian:squeeze'] + cmd_extras)
            assert False, 'system exit should have occurred'
        except SystemExit:
            pass
        try:
            rosdep_main(['search', 'nonexistent'] + cmd_extras + ['-i'])
            assert False, 'system exit should have occurred'
        except SystemExit:
            pass

    @patch('rosdep2.main.install_opener')
    @patch('rosdep2.main.build_opener')
    @patch('rosdep2.main.HTTPBasicAuthHandler')
    @patch('rosdep2.main.ProxyHandler')
    def test_proxy_detection(self, proxy, bah, build, install):
        with patch.dict('os.environ', {'http_proxy': 'something'}, clear=True):
            setup_proxy_opener()
            proxy.assert_called_with({'http': 'something'})
        with patch.dict('os.environ', {'https_proxy': 'somethings'}, clear=True):
            setup_proxy_opener()
            proxy.assert_called_with({'https': 'somethings'})

    def test_invalid_package_message(self):
        with fakeout() as b:
            test_package_dir = os.path.abspath(os.path.join(get_test_dir(), 'main', 'invalid_package_version'))
            with patch('rosdep2.main.sys.exit') as exit_mock:
                rosdep_main(['install', '--from-path', test_package_dir])
                exit_mock.assert_called_with(1)
            stdout, stderr = b
            output = stderr.getvalue().splitlines()
            assert len(output) >= 2
            assert test_package_dir in output[-2]
            assert 'Package version ":{version}" does not follow version conventions' in output[-1]



# Copyright (c) 2019, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import sys
from tempfile import TemporaryDirectory

from rosdep2.meta import MetaDatabase


def test_metadatabase_set_get():
    with TemporaryDirectory() as tmpdir:
        db = MetaDatabase(cache_dir=tmpdir)
        db.set('fruit', 'tomato')
        assert 'tomato' == db.get('fruit')


def test_metadatabase_get_none():
    with TemporaryDirectory() as tmpdir:
        db = MetaDatabase(cache_dir=tmpdir)
        assert db.get('fruit') is None


def test_metadatabase_get_default():
    with TemporaryDirectory() as tmpdir:
        db = MetaDatabase(cache_dir=tmpdir)
        assert db.get('fruit', default='foo') == 'foo'
        db.set('fruit', 'tomato')
        assert db.get('fruit', default='foo') != 'foo'


def test_metadatabase_get_mutate_get():
    with TemporaryDirectory() as tmpdir:
        db = MetaDatabase(cache_dir=tmpdir)
        mutable = [1, 2, 3]
        db.set('category', mutable)
        mutable.append(4)
        assert [1, 2, 3] == db.get('category')


def test_metadatabase_set_set_get():
    with TemporaryDirectory() as tmpdir:
        db = MetaDatabase(cache_dir=tmpdir)
        db.set('fruit', 'tomato')
        db.set('fruit', 'orange')
        assert 'orange' == db.get('fruit')


def test_metadatabase_set_load_from_disk_get():
    with TemporaryDirectory() as tmpdir:
        db1 = MetaDatabase(cache_dir=tmpdir)
        db1.set('fruit', 'apple')

        db2 = MetaDatabase(cache_dir=tmpdir)
        assert 'apple' == db2.get('fruit')
